from hape.logging import Logging
from hape.utils.naming_utils import NamingUtils
class StringUtils:
    
    logger = Logging.get_logger(__name__)
    
    @staticmethod
    def replace_name_case_placeholders(content: str, name: str, name_key_prefix: str) -> str:
        StringUtils.logger.debug(f"replace_name_case_placeholders(content, name: {name}, name_key_prefix: {name_key_prefix})")
        
        snake_case_key = f"{{{{{name_key_prefix}_snake_case}}}}"
        kebab_case_key = f"{{{{{name_key_prefix}_kebab_case}}}}"
        upper_snake_case_key = f"{{{{{name_key_prefix}_upper_snake_case}}}}"
        upper_kebab_case_key = f"{{{{{name_key_prefix}_upper_kebab_case}}}}"
        camel_case_key = f"{{{{{name_key_prefix}_camel_case}}}}"
        pascal_case_key = f"{{{{{name_key_prefix}_pascal_case}}}}"
        title_case_key = f"{{{{{name_key_prefix}_title_case}}}}"
        
        content = content.replace(snake_case_key, NamingUtils.convert_to_snake_case(name))
        content = content.replace(kebab_case_key, NamingUtils.convert_to_kebab_case(name))
        content = content.replace(upper_snake_case_key, NamingUtils.convert_to_upper_snake_case(name))
        content = content.replace(upper_kebab_case_key, NamingUtils.convert_to_upper_kebab_case(name))
        content = content.replace(camel_case_key, NamingUtils.convert_to_camel_case(name))
        content = content.replace(pascal_case_key, NamingUtils.convert_to_pascal_case(name))
        content = content.replace(title_case_key, NamingUtils.convert_to_title_case(name))
        return content


import os
import json
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import OperationalError
from sqlalchemy.sql import text
from hape.config import Config
from hape.logging import Logging
from hape.enums.environment_variables_enum import EnvironmentVariablesEnum
class HapeConfig:
    logger = Logging.get_logger('hape.config_db')
    _env_loaded = False
    _db_session = None
    _required_env_variables = []

    Config.set_required_env_variable("HAPE_GITLAB_TOKEN")
    Config.set_required_env_variable("HAPE_GITLAB_DOMAIN")
    Config.set_required_env_variable("HAPE_MARIADB_HOST")
    Config.set_required_env_variable("HAPE_MARIADB_USERNAME")
    Config.set_required_env_variable("HAPE_MARIADB_PASSWORD")
    Config.set_required_env_variable("HAPE_MARIADB_DATABASE")

    @staticmethod
    def check_variables():
        Config.logger.debug(f"check_variables()")
        for variable in Config._required_env_variables:
            Config._get_env_value(variable)

    @staticmethod
    def get_db_url(with_database=True):
        Config.logger.debug(f"get_db_url(with_database={with_database})")
        db_url = f"mysql+pymysql://{HapeConfig.get_mariadb_username()}:{HapeConfig.get_mariadb_password()}@{HapeConfig.get_mariadb_host()}"
        if with_database:
            db_url += f"/{HapeConfig.get_mariadb_database()}"
        return db_url
    

    @staticmethod
    def get_db_session(with_database=True) -> sessionmaker:
        Config.logger.debug(f"get_db_url()")
        if not HapeConfig._db_session:
            try:
                DATABASE_URL = HapeConfig.get_db_url(with_database)
                engine = create_engine(DATABASE_URL, echo=True)
                with engine.connect() as connection:
                    connection.execute(text("SELECT 1"))

                HapeConfig._db_session = sessionmaker(autocommit=False, autoflush=False, bind=engine)
                Config.logger.info("Database seassion created successfully.")
            except OperationalError:
                Config.logger.error("Error: Unable to connect to the database. Please check the configuration.")
                raise

        return HapeConfig._db_session()
    
    @staticmethod
    def drop_db():
        Config.logger.debug(f"drop_db()")
        engine = create_engine(HapeConfig.get_db_url(with_database=False), echo=True)
        with engine.connect() as connection:
            connection.execute(text(f"DROP DATABASE IF EXISTS {HapeConfig.get_mariadb_database()}"))

    @staticmethod
    def create_db():
        Config.logger.debug(f"create_db()")
        engine = create_engine(HapeConfig.get_db_url(with_database=False), echo=True)
        with engine.connect() as connection:
            connection.execute(text(f"CREATE DATABASE IF NOT EXISTS {HapeConfig.get_mariadb_database()}"))

    @staticmethod
    def reset_db():
        Config.logger.debug(f"reset_db()")
        HapeConfig.drop_db()
        HapeConfig.create_db()
            
    @staticmethod
    def get_gitlab_token():
        Config.logger.debug(f"get_gitlab_token()")
        return Config._get_env_value(EnvironmentVariablesEnum.HAPE_GITLAB_TOKEN)
    
    @staticmethod
    def get_gitlab_domain():
        Config.logger.debug(f"get_gitlab_domain()")
        return Config._get_env_value(EnvironmentVariablesEnum.HAPE_GITLAB_DOMAIN)
    
    @staticmethod
    def get_mariadb_host():
        Config.logger.debug(f"get_mariadb_host()")
        return Config._get_env_value(EnvironmentVariablesEnum.HAPE_MARIADB_HOST)

    @staticmethod
    def get_mariadb_username():
        Config.logger.debug(f"get_mariadb_username()")
        return Config._get_env_value(EnvironmentVariablesEnum.HAPE_MARIADB_USERNAME)

    @staticmethod
    def get_mariadb_password():
        Config.logger.debug(f"get_mariadb_password()")
        return Config._get_env_value(EnvironmentVariablesEnum.HAPE_MARIADB_PASSWORD)

    @staticmethod
    def get_mariadb_database():
        Config.logger.debug(f"get_mariadb_database()")
        return Config._get_env_value(EnvironmentVariablesEnum.HAPE_MARIADB_DATABASE)



import json
from hape.hape_cli.enums.crud_column_fk_on_delete import CrudColumnFkOnDeleteEnum
from hape.hape_cli.enums.crud_column_valid_properties import CrudColumnValidPropertiesEnum
from hape.hape_cli.enums.crud_column_valid_types import CrudColumnValidTypesEnum

class CrudModelSchema:
    
    valid_types = [valid_type.value for valid_type in CrudColumnValidTypesEnum]
    valid_properties = [valid_property.value for valid_property in CrudColumnValidPropertiesEnum]
    valid_foreign_key_on_delete = [valid_foreign_key_on_delete.value for valid_foreign_key_on_delete in CrudColumnFkOnDeleteEnum]

    _model_schema_json = """
{
    "valid_types": {{valid-types}},
    "valid_properties": {{valid-properties}},
    "valid_foreign_key_on_delete": {{valid-foreign-key-on-delete}},
    "foreign_key_syntax": "foreign-key(foreign-key-model.foreign-key-attribute, on-delete=foreign-key-on-delete)",
    
    "model-name": {
        "column_name": {"valid-type": ["valid-property"]},
        "id": {"valid-type": ["valid-property"]},
        "updated_at": {"valid-type": []},
        "name": {"valid-type": ["valid-property", "valid-property"]},
        "enabled": {"valid-type": []},
    }
    
    "example-model": {
        "id": {"int": ["primary"]},
        "updated_at": {"timestamp": []},
        "name": {"string": ["required", "unique"]},
        "enabled": {"bool": []}
    }
}
""".replace("{{valid-types}}", json.dumps(valid_types)) \
    .replace("{{valid-properties}}", json.dumps(valid_properties)) \
    .replace("{{valid-foreign-key-on-delete}}", json.dumps(valid_foreign_key_on_delete)) \
    .strip()

    _model_schema_yaml = """
valid_types: {{valid-types}}
valid_properties: {{valid-properties}}
valid_foreign_key_on_delete: {{valid-foreign-key-on-delete}}
foreign_key_syntax: "foreign-key(foreign-key-model.foreign-key-attribute, on-delete=foreign-key-on-delete)"

model-name:
  column_name:
    valid-type: 
      - valid-property
  id:
    valid-type: 
      - valid-property
  updated_at:
    valid-type: []
  name:
    valid-type: 
      - valid-property
      - valid-property
  enabled:
    valid-type: []

example-model:
  id:
    int: 
      - primary
  updated_at:
    timestamp: []
  name:
    string: 
      - required
      - unique
  enabled:
    bool: []
""".replace("{{valid-types}}", json.dumps(valid_types)) \
    .replace("{{valid-properties}}", json.dumps(valid_properties)) \
    .replace("{{valid-foreign-key-on-delete}}", json.dumps(valid_foreign_key_on_delete)) \
    .strip()



from hape.logging import Logging
from hape.hape_cli.interfaces.format_model import FormatModel
from hape.hape_cli.interfaces.format_controller import FormatController

class FormatArgumentParser:
    
    def __init__(self, format: str):
        self.logger = Logging.get_logger('hape.hape_cli.argument_parsers.format_argument_parser')
        if format not in FormatModel.supported_formats:
            self.logger.error(f"Invalid format: {format}. Supported formats are: {', '.join(FormatModel.supported_formats)}")
            exit(1)
        self.format = format
        self.COMMAND = format

    def create_subparser(self, subparsers):    
        self.logger.debug(f"create_subparser(subparsers)")
        format_parser = subparsers.add_parser(self.COMMAND, help=f"Commands related to {self.format.upper()} to generate model schema templates")
        format_parser_subparser = format_parser.add_subparsers(dest="action")

        format_parser = format_parser_subparser.add_parser("get", help=f"Get {self.format.upper()} templates and data related to the project")
        format_parser.add_argument("-m", "--model-schema", required=True, action="store_true", help=f"Template {self.format.upper()} schema of the model")

    def run_action(self, args):
        self.logger.debug(f"run_action(args)")
        if args.command != self.COMMAND:
            return
        if args.action == "get":
            FormatController(self.format).get()
        else:
            self.logger.error(f"Error: Invalid action {args.action} for {args.command}. Use `hape {args.command} --help` for more details.")
            exit(1)



import argparse
from importlib.metadata import version

from hape.logging import Logging
from hape.argument_parsers.playground_argument_parser import PlaygroundArgumentParser
from hape.argument_parsers.config_argument_parser import ConfigArgumentParser
from hape.argument_parsers.git_argument_parser import GitArgumentParser
from hape.argument_parsers.k8s_deployment_argument_parser import K8SDeploymentArgumentParser
from hape.argument_parsers.k8s_deployment_cost_argument_parser import K8SDeploymentCostArgumentParser

class MainArgumentParser:

    def __init__(self):
        self.logger = Logging.get_logger('hape.argument_parsers.main_argument_parser')

    def create_parser(self):
        self.logger.debug("create_parser()")
        parser = argparse.ArgumentParser(
            description="hape cli - streamline your development operations"
        )
        try:
            parser.add_argument("-v", "--version", action="version", version=version("hape"))
        except:
            parser.add_argument("-v", "--version", action="version", version="0.0.0")
        
        subparsers = parser.add_subparsers(dest="command")
        
        PlaygroundArgumentParser().create_subparser(subparsers)
        ConfigArgumentParser().create_subparser(subparsers)
        K8SDeploymentArgumentParser().create_subparser(subparsers)
        K8SDeploymentCostArgumentParser().create_subparser(subparsers)
        GitArgumentParser().create_subparser(subparsers)
        
        return parser
    
    def run_action(self, args):
        self.logger.debug("run_action()")
        if args.command == "play":
            PlaygroundArgumentParser().run_action(args)
        elif args.command == "config":
            ConfigArgumentParser().run_action(args)
        elif args.command == "k8s-deployment":
            K8SDeploymentArgumentParser().run_action(args)
        elif args.command == "k8s-deployment-cost":
            K8SDeploymentCostArgumentParser().run_action(args)
        elif args.command == "git":
            GitArgumentParser().run_action(args)
        else:
            self.logger.error(f"Invalid command {args.command}. Use `hape --help` for more details.")
            exit(1)



from hape.logging import Logging
from hape.utils.naming_utils import NamingUtils
from hape.hape_cli.enums.crud_column_fk_on_delete import CrudColumnFkOnDeleteEnum

class CrudColumnForeignKey:
    def __init__(self, foreign_key_property_value: str):
        self.logger = Logging.get_logger('hape.hape_cli.models.crud_column_foreign_key')
        
        self.foreign_key_table = ""
        self.foreign_key_column = ""
        self.foreign_key_on_delete = ""
        
        foreign_key_parts = foreign_key_property_value.split(",")
            
        foreign_key_table = NamingUtils.convert_to_snake_case(foreign_key_parts[0].split(".")[0])
        foreign_key_column = NamingUtils.convert_to_snake_case(foreign_key_parts[0].split(".")[1])
        
        self.foreign_key_table = foreign_key_table
        self.foreign_key_column = foreign_key_column
        
        foreign_key_on_delete = foreign_key_parts[1].split("=")[1]
        
        if foreign_key_on_delete == CrudColumnFkOnDeleteEnum.CASCADE.value:
            self.foreign_key_on_delete = "CASCADE"
        elif foreign_key_on_delete == CrudColumnFkOnDeleteEnum.SET_NULL.value:
            self.foreign_key_on_delete = "SET_NULL"
        elif foreign_key_on_delete == CrudColumnFkOnDeleteEnum.SET_DEFAULT.value:
            self.foreign_key_on_delete = "SET_DEFAULT"
        elif foreign_key_on_delete == CrudColumnFkOnDeleteEnum.RESTRICT.value:
            self.foreign_key_on_delete = "RESTRICT"
        elif foreign_key_on_delete == CrudColumnFkOnDeleteEnum.NO_ACTION.value:
            self.foreign_key_on_delete = "NO_ACTION"
        else:
            self.logger.error(f"Invalid on_delete value: {foreign_key_on_delete}")
            exit(1)
    


ARGUMENT_PARSER_TEMPLATE = """
from hape.logging import Logging
from hape.base.model_argument_parser import ModelArgumentParser
from {{project_name_snake_case}}.models.{{model_name_snake_case}}_model import {{model_name_pascal_case}}
from {{project_name_snake_case}}.controllers.{{model_name_snake_case}}_controller import {{model_name_pascal_case}}Controller

class {{model_name_pascal_case}}ArgumentParser(ModelArgumentParser):
    def __init__(self):
        super().__init__({{model_name_pascal_case}}, {{model_name_pascal_case}}Controller)
        self.logger = Logging.get_logger('{{project_name_snake_case}}.argument_parsers.{{model_name_snake_case}}_argument_parser')

    def extend_subparser(self):
        pass
    
    def extend_actions(self):
        pass

""".strip()


from hape.logging import Logging
from hape.hape_cli.controllers.crud_controller import CrudController

class CrudArgumentParser:
    def __init__(self):
        self.COMMAND = "crud"
        self.logger = Logging.get_logger('hape.argument_parsers.crud_argument_parser')

    def create_subparser(self, subparsers):    
        self.logger.debug(f"create_subparser(subparsers)")
        crud_parser = subparsers.add_parser(self.COMMAND, help="Commands related to CRUD operations")
        crud_parser_subparser = crud_parser.add_subparsers(dest="action")

        generate_parser = crud_parser_subparser.add_parser("generate", help="Generates a new CRUD operation")
        input_group = generate_parser.add_mutually_exclusive_group(required=True)
        input_group.add_argument("-n", "--name", help="Name of the model")
        input_group.add_argument("-j", "--json", nargs="?", const=True, help="Schema of the model in JSON format")
        input_group.add_argument("-y", "--yaml", nargs="?", const=True, help="Schema of the model in YAML format")
        
        delete_parser = crud_parser_subparser.add_parser("delete", help="Deletes a CRUD operation")
        delete_parser.add_argument("-n", "--name", required=True, help="Name of the model")
        
    def run_action(self, args):
        self.logger.debug(f"run_action(args.action: {args.action})")
        if args.command != self.COMMAND:
            return
        crud_controller = CrudController(
            args.name if "name" in args else None,
            args.json if "json" in args else None,
            args.yaml if "yaml" in args else None
        )
        
        if args.action == "generate":
            crud_controller.generate()
        elif args.action == "delete":
            crud_controller.delete()
        else:
            self.logger.error(f"Error: Invalid action {args.action} for {args.command}. Use `hape {args.command} --help` for more details.")
            exit(1)



MODEL_TEMPLATE = """
from hape.logging import Logging
from sqlalchemy import Column, Integer, String, Float, Boolean, BigInteger, ForeignKey, Index, Date, DateTime, TIMESTAMP, Text
from sqlalchemy.orm import relationship
from hape.base.model import Model

class {{model_name_pascal_case}}(Model):
    __tablename__ = '{{model_name_snake_case}}'
    logger = Logging.get_logger('{{project_name_snake_case}}.models.{{model_name_snake_case}}')
    
    {{model_columns}}
    {{model_relationships}}
    def __init__(self, **kwargs):
        filtered_kwargs = {key: kwargs[key] for key in self.__table__.columns.keys() if key in kwargs}
        super().__init__(**filtered_kwargs)
        for key, value in filtered_kwargs.items():
            setattr(self, key, value)
        self.logger = Logging.get_logger('{{project_name_snake_case}}.models.{{model_name_snake_case}}')
""".strip()


import argparse
from importlib.metadata import version
from hape.logging import Logging
from hape.hape_cli.argument_parsers.init_argument_parser import InitArgumentParser
from hape.hape_cli.argument_parsers.crud_argument_parser import CrudArgumentParser
from hape.hape_cli.argument_parsers.json_argument_parser import JsonArgumentParser
from hape.hape_cli.argument_parsers.yaml_argument_parser import YamlArgumentParser

class MainArgumentParser:
    
    def __init__(self):
        self.logger = Logging.get_logger('hape.hape_cli.argument_parsers.main_argument_parser')

    def create_parser(self):
        parser = argparse.ArgumentParser(description="HAPE Framework CLI")

        try:
            parser.add_argument("-v", "--version", action="version", version=version("hape"))
        except:
            parser.add_argument("-v", "--version", action="version", version="0.0.0")
        
        subparsers = parser.add_subparsers(dest="command")
        
        InitArgumentParser().create_subparser(subparsers)
        CrudArgumentParser().create_subparser(subparsers)
        JsonArgumentParser().create_subparser(subparsers)
        YamlArgumentParser().create_subparser(subparsers)
        
        return parser

    def run_action(self, args):

        if args.command == "init":
            InitArgumentParser().run_action(args)
        elif args.command == "crud":
            CrudArgumentParser().run_action(args)
        elif args.command == "json":
            JsonArgumentParser().run_action(args)
        elif args.command == "yaml":
            YamlArgumentParser().run_action(args)
        else:
            self.logger.error(f"Error: Main Parser Invalid command {args.command}")
            exit(1)



MIGRATION_TEMPLATE = """
from alembic import op
import sqlalchemy as sa

revision = '{{migration_counter}}'
down_revision = None
branch_labels = None
depends_on = None

def upgrade():
    {{model_table_creation_statements}}
def downgrade():
    {{model_table_drop_statements}}
""".strip()

MIGRATION_MODEL_TABLE_CREATION_TEMPLATE = """
    op.create_table(
        '{{model_name_snake_case}}',
        {{migration_columns}}
    )
""".strip()

MIGRATION_MODEL_TABLE_DROP_TEMPLATE = """
    op.drop_table('{{model_name_snake_case}}')
""".strip()



import enum

class EnvironmentVariablesEnum(enum.Enum):
    HAPE_LOG_FILE = "HAPE_LOG_FILE"
    HAPE_LOG_LEVEL = "HAPE_LOG_LEVEL"
    HAPE_LOG_ROTATE_EVERY_RUN = "HAPE_LOG_ROTATE_EVERY_RUN"
    HAPE_MARIADB_HOST = "HAPE_MARIADB_HOST"
    HAPE_MARIADB_USERNAME = "HAPE_MARIADB_USERNAME"
    HAPE_MARIADB_PASSWORD = "HAPE_MARIADB_PASSWORD"
    HAPE_MARIADB_DATABASE = "HAPE_MARIADB_DATABASE"
    HAPE_GITLAB_DOMAIN = "HAPE_GITLAB_DOMAIN"
    HAPE_GITLAB_TOKEN = "HAPE_GITLAB_TOKEN"



from hape.hape_cli.enums.crud_column_valid_properties import CrudColumnValidPropertiesEnum
from hape.hape_cli.models.crud_column_foreign_key import CrudColumnForeignKey

class CrudColumnProperty:
    def __init__(self, property: CrudColumnValidPropertiesEnum, foreign_key_value: str = None):
        self.property: CrudColumnValidPropertiesEnum = CrudColumnValidPropertiesEnum(property)
        self.foreign_key: CrudColumnForeignKey = None
        
        if property == CrudColumnValidPropertiesEnum.FOREIGN_KEY and foreign_key_value:
            self.foreign_key = CrudColumnForeignKey(foreign_key_value)



from hape.logging import Logging
from hape.base.model_argument_parser import ModelArgumentParser
from hape.models.k8s_deployment_model import K8SDeployment
from hape.controllers.k8s_deployment_controller import K8SDeploymentController

class K8SDeploymentArgumentParser(ModelArgumentParser):
    def __init__(self):
        super().__init__(K8SDeployment, K8SDeploymentController)
        self.logger = Logging.get_logger('hape.argument_parsers.k8s_deployment_argument_parser')

    def extend_subparser(self):
        pass
    
    def extend_actions(self):
        pass


from hape.logging import Logging
from sqlalchemy import Column, Integer, String, Float, Boolean, BigInteger, ForeignKey, Index, Date, DateTime, TIMESTAMP, Text
from sqlalchemy.orm import relationship
from hape.base.model import Model

class K8SDeployment(Model):
    __tablename__ = 'k8s_deployment'
    logger = Logging.get_logger('hape.models.k8s_deployment')
    
    id = Column(Integer, primary_key=True, autoincrement=True)
    service_name = Column(String(255), nullable=False)
    pod_cpu = Column(String(255), nullable=False)
    pod_ram = Column(String(255), nullable=False)
    autoscaling = Column(Boolean, nullable=False)
    min_replicas = Column(Integer, nullable=True)
    max_replicas = Column(Integer, nullable=True)
    current_replicas = Column(Integer, nullable=False)
    
    def __init__(self, **kwargs):
        filtered_kwargs = {key: kwargs[key] for key in self.__table__.columns.keys() if key in kwargs}
        super().__init__(**filtered_kwargs)
        for key, value in filtered_kwargs.items():
            setattr(self, key, value)
        self.logger = Logging.get_logger('hape.models.k8s_deployment')


import json
from datetime import datetime
from hape.config import Config
from hape.hape_config import HapeConfig
from hape.services.gitlab_service import GitlabService
from hape.services.file_service import FileService
from hape.hape_cli.models.crud_model import Crud
from hape.hape_cli.models.json_model import Json
from hape.hape_cli.models.yaml_model import Yaml
from hape.utils.naming_utils import NamingUtils

class Playground:

    @classmethod
    def main(self):
        playground = Playground()
        playground.play()
        
    def generate_gitlab_changes_report(self):
        gitlab = GitlabService()
        start_date = datetime(2025, 2, 3)
        end_date = datetime(2025, 2, 5)
        gitlab.generate_csv_changes_in_cicd_repos(
            group_id=178,
            start_date=start_date,
            end_date=end_date,
            output_file="/Users/hazemataya/Desktop/workspace/innodp/playground/test.csv",
            file_regex=r".*values.*.yaml"
        )
        
    def play_with_crud(self):
        try:
            Crud(
                project_name="hape",
                model_name="k8s-deployment"
            ).delete()
        except Exception as e:
            print(f"Error deleting k8s-deployment: {e}")
            raise e

        try:
            Crud(
                project_name="hape",
                model_name="k8s-deployment-cost"
            ).delete()
        except Exception as e:
            print(f"Error deleting k8s-deployment-cost: {e}")
            raise e
                    
        Crud(
            project_name="hape",
            schemas={   
            "k8s-deployment": {
                "id": {"int": ["primary", "autoincrement"]},
                "service-name": {"string": ["required"]},
                "pod-cpu": {"string": ["required"]},
                "pod-ram": {"string": ["required"]},
                "autoscaling": {"bool": ["required"]},
                "min-replicas": {"int": ["nullable"]},
                "max-replicas": {"int": ["nullable"]},
                "current-replicas": {"int": ["required"]},
            },
            "k8s-deployment-cost": {
                "id": {"int": ["primary", "autoincrement"]},
                "k8s-deployment-id": {"int": ["required", "foreign-key(k8s-deployment.id, on-delete=cascade)"]},
                "pod-cost": {"string": ["required"]},
                "number-of-pods": {"int": ["required"]},
                "total-cost": {"float": ["required"]}
                }
            }
        ).generate()

        # def play_with_k8s_deployment(self):
        #     k8s_deployment = K8SDeployment(
        #         service_name="test",
        #         pod_cpu="1",
        #         pod_ram="1",
        #         autoscaling=True,
        #         min_replicas=1,
        #         max_replicas=10,
        #         current_replicas=1
        #     )
        #     print(k8s_deployment.json())
        #     print(k8s_deployment.validate())
        #     exit()
        #     k8s_deployment.save()
            
        #     k8s_deployment_cost = K8SDeploymentCost(
        #         k8s_deployment_id=k8s_deployment.id,
        #         pod_cost="100",
        #         number_of_pods=1,
        #         total_cost=100
        #     )
        #     k8s_deployment_cost.save()
            
        #     print(K8SDeployment.list_to_json(K8SDeployment.get_all()))
        #     print(K8SDeploymentCost.list_to_json(K8SDeploymentCost.get_all()))


    def play(self):
        # HapeConfig.reset_db()
        # self.play_with_crud()
        var = "my-test-var"
        print("original: " + var)
        print("snake_case: " + NamingUtils.convert_to_snake_case(var))
        print("kebab_case: " + NamingUtils.convert_to_kebab_case(var))
        print("upper_snake_case: " + NamingUtils.convert_to_upper_snake_case(var))
        print("upper_kebab_case: " + NamingUtils.convert_to_upper_kebab_case(var))
        print("camel_case: " + NamingUtils.convert_to_camel_case(var))
        print("pascal_case: " + NamingUtils.convert_to_pascal_case(var))
        print("title_case: " + NamingUtils.convert_to_title_case(var))



import json
import yaml
from hape.logging import Logging
from hape.hape_cli.models.crud_model_schema import CrudModelSchema

class FormatModel:
    
    supported_formats = ["json", "yaml"]
    
    def __init__(self, format: str):
        self.logger = Logging.get_logger('hape.hape_cli.interfaces.format_model')
        self.schema = None
        self.format = format
    
    def load(self, schema: str):
        if self.format == "json":
            self.schema = json.loads(schema)
        elif self.format == "yaml":
            self.schema = yaml.safe_load(schema)
        else:
            self.logger.error(f"Invalid format: {self.format}")
            exit(1)
        return self.schema
    
    def get(self):
        if self.format == "json":
            print(CrudModelSchema._model_schema_json)
        elif self.format == "yaml":
            print(CrudModelSchema._model_schema_yaml)
        else:
            self.logger.error(f"Invalid format: {self.format}. Supported formats are: json, yaml")
            exit(1)
        


import enum

class CrudColumnValidPropertiesEnum(enum.Enum):
    NULLABLE = "nullable"
    REQUIRED = "required"
    UNIQUE = "unique"
    PRIMARY = "primary"
    AUTOINCREMENT = "autoincrement"
    FOREIGN_KEY = "foreign-key"
    INDEX = "index"



import json
import yaml
import os

from hape.logging import Logging
from hape.hape_cli.models.crud_model import Crud
from hape.services.file_service import FileService

class CrudController:

    # name is added for the --delete command, to be able to pass the name.
    def __init__(self, name, schema_json, schema_yaml):
        self.logger = Logging.get_logger('hape.hape_cli.controllers.crud_controller')
        file_service = FileService()
        
        schema = None
        
        is_json = bool(schema_json)
        is_yaml = bool(schema_yaml)    
        is_json_string = is_json and isinstance(schema_json, str)
        is_yaml_string = is_yaml and isinstance(schema_yaml, str)
        is_json_file = is_json_string and schema_json.endswith('.json')
        is_yaml_file = is_yaml_string and schema_yaml.endswith('.yaml')
        is_json_default = is_json and isinstance(schema_json, bool)
        is_yaml_default = is_yaml and isinstance(schema_yaml, bool)
        self.logger.debug(f"is_json_string: {is_json_string}")
        self.logger.debug(f"is_yaml_string: {is_yaml_string}")
        self.logger.debug(f"is_json_file: {is_json_file}")
        self.logger.debug(f"is_yaml_file: {is_yaml_file}")
        self.logger.debug(f"is_json_default: {is_json_default}")
        self.logger.debug(f"is_yaml_default: {is_yaml_default}")

        if is_json_string:
            try:
                schema = json.loads(schema_json)
            except e:
                self.logger.error(f"Invalid JSON schema: {schema_json}")
                exit(1)
        elif is_yaml_string:
            try:
                schema = yaml.safe_load(schema_yaml)
            except yaml.YAMLError as e:
                self.logger.error(f"Invalid YAML schema: {schema_yaml}")
                exit(1)
        if is_json_file:
            if not FileService().file_exists(schema_json):
                self.logger.error(f"JSON file not found at {schema_json}")
                exit(1)
            schema = file_service.read_json_file(schema_json)
        elif is_yaml_file:
            if not FileService().file_exists(schema_yaml):
                self.logger.error(f"YAML file not found at {schema_yaml}")
                exit(1)
            schema = file_service.read_yaml_file(schema_yaml)
        elif is_json_default:
            if not FileService().file_exists(Crud.draft_json_file_path):
                self.logger.error(f"Draft JSON file not found at {Crud.draft_json_file_path}")
                exit(1)
            schema = file_service.read_json_file(Crud.draft_json_file_path)
        elif is_yaml_default:
            if not FileService().file_exists(Crud.draft_yaml_file_path):
                self.logger.error(f"Draft YAML file not found at {Crud.draft_yaml_file_path}")
                exit(1)
            schema = file_service.read_yaml_file(Crud.draft_yaml_file_path)
            
        self.crud = Crud(os.path.basename(os.getcwd()), name, schema)
        self.crud.validate()
    
    def generate(self):
        self.crud.validate_schemas()
        self.crud.generate()
        
    def delete(self):
        self.crud.delete()



from hape.hape_config import HapeConfig
from hape.logging import Logging
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session
import json

Base = declarative_base()

class Model(Base):
    __abstract__ = True
    
    __required_fields = {}
    __field_types = {}

    logger = Logging.get_logger('hape.base.model')

    def __init__(self, **kwargs):
        self.logger = Logging.get_logger('hape.base.model')

    @classmethod
    def initialize_from_sqlalchemy(cls, sqlalchemy_base_model):
        cls.logger.debug(f"initialize_from_sqlalchemy({sqlalchemy_base_model})")
        cls.__required_fields = {
            column.name: not column.nullable for column in sqlalchemy_base_model.__table__.columns
        }
        cls.__field_types = {
            column.name: column.type.python_type for column in sqlalchemy_base_model.__table__.columns
        }
    
    def validate(self):
        self.logger.debug(f"validate()")
        for field, is_required in self.__required_fields.items():
            if is_required and field not in self.__dict__ or self.__dict__[field] is None:
                if field not in ['id', 'created_at']:
                    return False
        for field, field_type in self.__field_types.items():
            if field in self.__dict__ and not isinstance(self.__dict__[field], field_type):
                print("invalid!")
                return False
        return True
    
    def dict(self):
        return {column.name: getattr(self, column.name) for column in self.__table__.columns}

    def json(self, pretty=True):
        return json.dumps(self.dict(), indent=4) if pretty else json.dumps(self.dict())
    
    @classmethod
    def list_to_json(cls, objects, pretty=True):
        return json.dumps([obj.dict() for obj in objects], indent=4) if pretty else json.dumps([obj.dict() for obj in objects])

    @classmethod
    def _get_session(cls) -> Session:
        return HapeConfig.get_db_session()

    def save(self):
        self.logger.debug(f"save({self.json(pretty=False)})")
        session = Model._get_session() 
        exit_status = True
        try:
            session.add(self)
            session.commit()
            session.refresh(self)
        except Exception:
            exit_status = False
            session.rollback()
            print("---rollback---")
        finally:
            session.close()
            print("---close---")
            return exit_status

    @classmethod
    def get(cls, **filters):
        cls.logger.debug(f"get({filters})")
        session = cls._get_session()
        try:
            query = session.query(cls)
            for key, value in filters.items():
                if isinstance(value, list):
                    query = query.filter(getattr(cls, key).in_(value))
                else:
                    query = query.filter(getattr(cls, key) == value)
            return query.first()
        except Exception:
            return None
        finally:
            session.close()

    @classmethod
    def get_all(cls, **filters):
        cls.logger.debug(f"get_all({filters})")
        session = cls._get_session()
        try:
            query = session.query(cls)
            for key, value in filters.items():
                if isinstance(value, list):
                    query = query.filter(getattr(cls, key).in_(value))
                else:
                    query = query.filter(getattr(cls, key) == value)
            return query.all()
        except Exception:
            return []
        finally:
            session.close()

    def delete(self):
        self.logger.debug(f"delete()")
        session = Model._get_session()
        try:
            session.delete(self)
            session.commit()
        except Exception:
            session.rollback()
        finally:
            session.close()

    @classmethod
    def delete_all(cls, **filters):
        cls.logger.debug(f"delete_all({filters})")
        session = cls._get_session()
        try:
            query = session.query(cls)
            for key, value in filters.items():
                if isinstance(value, list):
                    query = query.filter(getattr(cls, key).in_(value))
                else:
                    query = query.filter(getattr(cls, key) == value)
            deleted_count = query.delete(synchronize_session=False)
            session.commit()
            return deleted_count
        except Exception:
            session.rollback()
            return 0
        finally:
            session.close()



from hape.logging import Logging
from sqlalchemy import Column, Integer, String, Float, Boolean, BigInteger, ForeignKey, Index, Date, DateTime, TIMESTAMP, Text
from sqlalchemy.orm import relationship
from hape.base.model import Model

class K8SDeploymentCost(Model):
    __tablename__ = 'k8s_deployment_cost'
    logger = Logging.get_logger('hape.models.k8s_deployment_cost')
    
    id = Column(Integer, primary_key=True, autoincrement=True)
    k8s_deployment_id = Column(Integer, ForeignKey('k8s_deployment.id', ondelete='CASCADE'), nullable=False)
    pod_cost = Column(String(255), nullable=False)
    number_of_pods = Column(Integer, nullable=False)
    total_cost = Column(Float, nullable=False)

    relationship('K8SDeployment', back_populates='k8s_deployments')

    def __init__(self, **kwargs):
        filtered_kwargs = {key: kwargs[key] for key in self.__table__.columns.keys() if key in kwargs}
        super().__init__(**filtered_kwargs)
        for key, value in filtered_kwargs.items():
            setattr(self, key, value)
        self.logger = Logging.get_logger('hape.models.k8s_deployment_cost')


from hape.logging import Logging
from hape.hape_cli.interfaces.format_controller import FormatController

class YamlController(FormatController):

    def __init__(self):
        super().__init__("yaml")
        self.logger = Logging.get_logger('hape.hape_cli.controllers.yaml_controller')


from hape.logging import Logging

import re
from typing import List, Optional
from datetime import datetime
from gitlab import Gitlab
from gitlab.v4.objects import Group, Project, GroupProject, ProjectCommit
from hape.hape_config import HapeConfig
from hape.services.file_service import FileService

class GitlabService:
    def __init__(self):
        self.logger = Logging.get_logger('hape.services.gitlab_service')
        gitlab_domain = HapeConfig.get_gitlab_domain()
        gitlab_token = HapeConfig.get_gitlab_token()
        self.client = Gitlab(f"https://{gitlab_domain}", private_token=gitlab_token)
        self._per_page = 100
    
    def _get_project(self, id: int) -> Project:
        return self.client.projects.get(id)
    
    def _get_projects(self, ids: list[int]) -> List[Project]:
        return [self.client.projects.get(id) for id in ids]
    
    def _get_group_projects(self, group_id: int, recursive: bool = False, search: Optional[str] = None) -> List[GroupProject]:
        group: Group = self.client.groups.get(group_id)
        return group.projects.list(get_all=True, per_page=self._per_page, include_subgroups=recursive, search=search)        
    
    def _get_project_commits(self, project: Project, start_date: Optional[datetime] = None, end_date: Optional[datetime] = None) -> List[ProjectCommit]:
        params = {"get_all": True}
        if start_date:
            params["since"] = start_date.isoformat()
        if end_date:
            params["until"] = end_date.isoformat()
        
        return project.commits.list(**params)
    
    def generate_csv_changes_in_cicd_repos(self, group_id: int, output_file: str, start_date: Optional[datetime] = None, end_date: Optional[datetime] = None, file_regex: Optional[str] = None):
        self.logger.debug(f"generate_csv_changes_in_cicd_repos(group_id: int: {group_id: int}, output_file: str: {output_file: str}, start_date: Optional[datetime] : {start_date: Optional[datetime] }, end_date: Optional[datetime] : {end_date: Optional[datetime] }, file_regex: Optional[str] : {file_regex: Optional[str] })")
        
        group_projects = self._get_group_projects(group_id, recursive=True, search="cicd-")
        project_ids = [project.id for project in group_projects]
        projects = self._get_projects(project_ids)

        all_commits = []
        for project in projects:
            project_path = project.path.replace("cicd-", "")
            commits = self._get_project_commits(project, start_date, end_date)
            for commit in commits:
                if commit.author_name == f"{project_path}-ci":
                    continue

                diff = commit.diff()
                for change in diff:
                    if file_regex and re.match(file_regex, change['new_path']):
                        all_commits.append({
                            "project_name": project.name,
                            "commit_date": commit.committed_date,
                            "commit_author": commit.author_name,
                            "commit": commit.id,
                            "web_url": commit.web_url
                        })

        FileService().write_csv_file(output_file, all_commits)
        print(f"CSV report generated: {output_file}")



from hape.logging import Logging
from hape.utils.naming_utils import NamingUtils
from typing import List
from hape.hape_cli.enums.crud_column_valid_properties import CrudColumnValidPropertiesEnum
from hape.hape_cli.enums.crud_column_valid_types import CrudColumnValidTypesEnum
from hape.hape_cli.enums.crud_column_fk_on_delete import CrudColumnFkOnDeleteEnum
from hape.hape_cli.models.crud_column_property import CrudColumnProperty

class CrudColumn:
    def __init__(self, name: str, type: str, properties: List[str]):
        self.logger = Logging.get_logger('hape.hape_cli.models.crud_column')
        
        self.name: str = name
        self.type: CrudColumnValidTypesEnum = CrudColumnValidTypesEnum(type)
        
        self.crud_column_properties: List[CrudColumnProperty] = self._parse_properties(properties)
        
        self.is_id = name == 'id'
        self.is_primary = CrudColumnValidPropertiesEnum.PRIMARY.value in properties
        self.is_autoincrement = CrudColumnValidPropertiesEnum.AUTOINCREMENT.value in properties
        self.is_unique = CrudColumnValidPropertiesEnum.UNIQUE.value in properties
        self.is_index = CrudColumnValidPropertiesEnum.INDEX.value in properties
        self.is_nullable = CrudColumnValidPropertiesEnum.NULLABLE.value in properties
        self.is_required = CrudColumnValidPropertiesEnum.REQUIRED.value in properties
         
    def _parse_properties(self, properties: List[str]) -> List[CrudColumnProperty]:
        self.logger.debug(f"_parse_properties()")
        
        crud_column_properties = []
        
        for property in properties:
            crud_column_property = None
            if property.startswith("foreign-key"):
                foreign_key_value = property.split("(")[1].split(")")[0]
                crud_column_property = CrudColumnProperty(
                    CrudColumnValidPropertiesEnum.FOREIGN_KEY,
                    foreign_key_value
                )
            else:
                crud_column_property = CrudColumnProperty(property)
                
            crud_column_properties.append(crud_column_property)
        
        if not crud_column_properties or (
            CrudColumnValidPropertiesEnum.NULLABLE.value not in properties 
            and CrudColumnValidPropertiesEnum.REQUIRED.value not in properties
            and CrudColumnValidPropertiesEnum.PRIMARY.value not in properties
        ):
            crud_column_properties.append(CrudColumnProperty(CrudColumnValidPropertiesEnum.NULLABLE.value))
        
        return crud_column_properties



from hape.logging import Logging
from hape.hape_cli.interfaces.format_model import FormatModel

class Yaml(FormatModel):
    def __init__(self):
        super().__init__("yaml")
        self.logger = Logging.get_logger('hape.hape_cli.models.yaml_model')



class NamingUtils:

    def _make_first_letter_upper(name: str) -> str:
        return name[:1].upper() + name[1:] if name else name

    def _make_first_letter_lower(name: str) -> str:
        return name[:1].lower() + name[1:] if name else name

    def convert_to_snake_case(name: str) -> str:
        return name.replace("-", "_").replace(" ", "_")
    
    def convert_to_kebab_case(name: str) -> str:
        return name.replace("_", "-").replace(" ", "-")
    
    def convert_to_upper_snake_case(name: str) -> str:
        return name.upper().replace("-", "_").replace(" ", "_")
    
    def convert_to_upper_kebab_case(name: str) -> str:
        return name.upper().replace("_", "-").replace(" ", "-")
    
    def convert_to_camel_case(name: str) -> str:
        return NamingUtils._make_first_letter_lower(name.title().replace("_", "").replace("-", ""))
    
    def convert_to_pascal_case(name: str) -> str:
        return NamingUtils._make_first_letter_upper(name.title().replace("_", "").replace("-", ""))
    
    def convert_to_title_case(name: str) -> str:
        return name.title().replace("_", " ").replace("-", " ")
    
    


from hape.logging import Logging
from hape.config import Config

class ConfigArgumentParser:

    def __init__(self):
        self.logger = Logging.get_logger('hape.argument_parsers.config_argument_parser')

    def create_subparser(self, subparsers):    
        self.logger.debug(f"create_subparser(subparsers)")
        config_parser = subparsers.add_parser("config", help="Commands related to the configurations of the cli")
        
        config_subparser = config_parser.add_subparsers(dest="action")
        check_parser = config_subparser.add_parser("check", help="Check if required configurations for the cli to work are set")

    def run_action(self, args):
        self.logger.debug(f"run_action(args)")
        if args.command != "config":
            return
        
        if args.action == "check":
            Config.check_variables(Config._required_env_variables)
            self.logger.info("Configurations are set correctly.")
        else:
            self.logger.error(f"Error: Invalid action {args.action} for {args.command}. Use `hape {args.command} --help` for more details.")
            exit(1)



import re
import os
from typing import List, Literal
from alembic import command
from alembic.config import Config
from alembic.script import ScriptDirectory
from hape.logging import Logging
from hape.services.file_service import FileService
from hape.hape_cli.enums.crud_actions import CrudActionsEnum
from hape.hape_cli.crud_templates.argument_parser_template import ARGUMENT_PARSER_TEMPLATE
from hape.hape_cli.crud_templates.controller_template import CONTROLLER_TEMPLATE
from hape.hape_cli.crud_templates.migration_template import MIGRATION_TEMPLATE, MIGRATION_MODEL_TABLE_CREATION_TEMPLATE, MIGRATION_MODEL_TABLE_DROP_TEMPLATE
from hape.hape_cli.crud_templates.model_template import MODEL_TEMPLATE
from hape.utils.naming_utils import NamingUtils
from hape.utils.string_utils import StringUtils
from hape.hape_cli.models.crud_column_parser import CrudColumnParser
from hape.hape_cli.models.crud_column import CrudColumn
from hape.hape_cli.models.crud_model_schema import CrudModelSchema

class Crud:
    logger = Logging.get_logger('hape.hape_cli.models.crud_model')
    project_name = os.path.basename(os.getcwd())
    project_name_snake_case = "hape" if project_name == "hape-framework" else NamingUtils.convert_to_snake_case(project_name)
    
    draft_json_file_path = os.path.join(project_name_snake_case, "migrations", "json", "draft.json")
    draft_yaml_file_path = os.path.join(project_name_snake_case, "migrations", "yaml", "draft.yaml")
    
    def __init__(self, project_name: str, model_name: str = None, schemas: dict[str, dict] = None):
        self.logger = Logging.get_logger('hape.hape_cli.models.crud_model')
        self.file_service = FileService()
        self.action = CrudActionsEnum.GENERATE
        
        self.model_count = 0
        if schemas:
            self.model_count = len(schemas)
        elif model_name:
            self.model_count = 1
        else:
            self.logger.error("Model name or schema is required to create a CRUD object.")
            exit(1)
        
        self.schemas: dict[str, dict] = {}
        self.model_names: list[str] = []
        self.model_names_snake_case: dict[str, str] = {}
        if schemas:
            self.schemas = schemas
            for model_name, _ in schemas.items():
                self.model_names.append(model_name)
                self.model_names_snake_case[model_name] = NamingUtils.convert_to_snake_case(model_name)
        elif model_name:
            self.model_names.append(model_name)
            self.model_names_snake_case[model_name] = NamingUtils.convert_to_snake_case(model_name)
        else:
            self.logger.error("Model name or schema is required to create a CRUD object.")
            exit(1)
        
        self.project_name = project_name
        self.source_code_path = NamingUtils.convert_to_snake_case(project_name)
        if self.project_name == "hape-framework":
            self.source_code_path = "hape"
                
        self.argument_parser_paths: dict[str, str] = {}
        for model_name in self.model_names:
            self.argument_parser_paths[model_name] = os.path.join(self.source_code_path, "argument_parsers", f"{self.model_names_snake_case[model_name]}_argument_parser.py")
        self.controller_paths: dict[str, str] = {}
        for model_name in self.model_names:
            self.controller_paths[model_name] = os.path.join(self.source_code_path, "controllers", f"{self.model_names_snake_case[model_name]}_controller.py")
        self.model_paths: dict[str, str] = {}
        for model_name in self.model_names:
            self.model_paths[model_name] = os.path.join(self.source_code_path, "models", f"{self.model_names_snake_case[model_name]}_model.py")
            
        self.argument_parser_contents: dict[str, str] = {}
        self.controller_contents: dict[str, str] = {}
        self.model_contents: dict[str, str] = {}
        self.migration_content: str = ""
        
        self.argument_parser_generated: dict[str, bool] = {}
        self.controller_generated: dict[str, bool] = {}
        self.model_generated: dict[str, bool] = {}
        self.migration_generated: bool = False
        self.should_generate_migration: bool = False
        
        self.crud_columns: dict[str, List[CrudColumn]] = {}
        self.crud_column_parsers: dict[str, List[CrudColumnParser]] = {}
        
        self.is_default_migration_counter: bool = False
        self.migration_counter_digits: int = 6
        self.migration_counter: str = "000001"
        self.migration_columns: str = ""
        self.migration_path = os.path.join(self.source_code_path, "migrations", "versions", f"{self.migration_counter}_migration.py")
        self.migration_json_directory = os.path.join(self.source_code_path, "migrations", "json", f"{self.migration_counter}_migration.json")
        self.migration_yaml_directory = os.path.join(self.source_code_path, "migrations", "yaml", f"{self.migration_counter}_migration.yaml")
        self.alembic_config_path = os.path.join(os.getcwd(), "alembic.ini")
        
        self.models_tables_creation_statements = ""
        self.models_tables_drop_statements = ""
        
        if self.schemas:
            self._init_object_model_dictionaries()
        
    def _init_object_model_dictionaries(self):
        self.logger.debug(f"_init_object_model_dictionaries()")
        if not self.schemas:
            self.logger.warning("Schema is required")
            return
        
        for model_name, _ in self.schemas.items():
            self.argument_parser_contents[model_name] = ""
            self.controller_contents[model_name] = ""
            self.model_contents[model_name] = ""
            self.argument_parser_generated[model_name] = False
            self.controller_generated[model_name] = False
            self.model_generated[model_name] = False
            self.migration_generated = False
            self.crud_columns[model_name] = []
            self.crud_column_parsers[model_name] = []
        
        for model_name, columns in self.schemas.items():
            for column_name, column_type_and_properties in columns.items():
                crud_column_name = column_name
                crud_column_type = list(column_type_and_properties.keys())[0]
                crud_column_properties = list(column_type_and_properties.values())[0]    
                crud_column = CrudColumn(
                    crud_column_name,
                    crud_column_type,
                    crud_column_properties
                )
                self.crud_column_parsers[model_name].append(CrudColumnParser(crud_column))
                
    def validate(self):
        self.logger.debug(f"validate()")
        if not self.model_names:
            self.logger.error("Model name is required")
            exit(1)
        for model_name in self.model_names:
            if not re.match(r'^[a-z0-9]+(-[a-z0-9]+)*$', model_name):
                self.logger.error(f"Error: Model name '{model_name}' must contain only lowercase letters, numbers, and use '-' as a separator.")
                exit(1)
    
    def validate_schemas(self):
        self.logger.debug(f"validate_schemas()")
        if not self.schemas:
            self.logger.error("Schema is required")
            exit(1)
            
        self._validate_schemas_structure()
        
    def _validate_schemas_structure(self):
        self.logger.debug(f"_validate_schemas_structure()")
        if not self.schemas:
            self.logger.error("Schema is required")
            exit(1)
        if not isinstance(self.schemas, dict):
            self.logger.error(f"Schema must be a dictionary, but got {type(self.schemas)}: {self.schemas}")
            exit(1)
        if not self.schemas.keys():
            self.logger.error("Schema must be a dictionary have at least one key")
            exit(1)
        for model_name, columns in self.schemas.items():
            if not isinstance(model_name, str):
                self.logger.error(f"Model name must be a string, but got {type(model_name)}: {model_name}")
                exit(1)
            if not re.match(r'^[a-z0-9]+(-[a-z0-9]+)*$', model_name):
                self.logger.error(f"Model name '{model_name}' must contain only lowercase letters, numbers, and use '-' as a separator.")
                exit(1)
            for column_name, column_type_and_properties in columns.items():
                if not isinstance(column_name, str):
                    self.logger.error(f"Column name must be a string, but got {type(column_name)}: {column_name}")
                    exit(1)
                if not re.match(r'^[a-z0-9]+(-[a-z0-9]+)*$', column_name):
                    self.logger.error(f"Column name '{column_name}' must contain only lowercase letters, numbers, and use '-' as a separator.")
                    exit(1)
                if not isinstance(column_type_and_properties, dict):
                    self.logger.error(f"Each column must have be a dictionary, but got {type(column_type_and_properties)}: {column_type_and_properties}")
                    exit(1)
                
                column_type = list(column_type_and_properties.keys())[0]
                column_properties = list(column_type_and_properties.values())[0]
            
                if not isinstance(column_type, str):
                    self.logger.error(f"Each column must have a type, but got {type(column_type)}: {column_type}")
                    exit(1)
                if column_type not in CrudModelSchema.valid_types:
                    self.logger.error(f"Invalid column type '{column_type}'. Must be one of {CrudModelSchema.valid_types}")
                    exit(1)
                if not isinstance(column_properties, list):
                    self.logger.error(f"Each column must have a list of properties or empty list, but got {type(column_properties)}: {column_properties}")
                    exit(1)
                for column_property in column_properties:
                    if not isinstance(column_property, str):
                        self.logger.error("Each column property must be a string")
                        exit(1)
                    if column_property not in CrudModelSchema.valid_properties and not column_property.startswith("foreign-key"):
                        self.logger.error(f"Invalid column property '{column_property}'. Must be one of {CrudModelSchema.valid_properties}")
                        exit(1)
    
    def _get_orm_columns(self, model_name):
        self.logger.debug(f"_get_orm_columns(model_name: {model_name})")
        if not self.schemas:
            self.logger.warning(f"Schema is required")
            return ""

        parsed_orm_columns = ""
        splitter = "\n    "
        for crud_column_parser in self.crud_column_parsers[model_name]:
            parsed_orm_columns += crud_column_parser.parsed_orm_column + splitter

        return parsed_orm_columns.rstrip(splitter)
    
    def _get_orm_relationships(self, model_name):
        self.logger.debug(f"_get_orm_relationships(model_name: {model_name})")
        if not self.schemas:
            self.logger.warning("Schema is required")
            return ""
        
        parsed_orm_relationships = ""
        splitter = ",\n    "
        for crud_column_parser in self.crud_column_parsers[model_name]:
            if crud_column_parser.orm_relationships:
                parsed_orm_relationships += crud_column_parser.orm_relationships + splitter
        
        return parsed_orm_relationships.rstrip(splitter)
    
    def _generate_content_model(self, model_name):
        self.logger.debug(f"_generate_content_model(model_name: {model_name})")
        if self.file_service.file_exists(self.model_paths[model_name]):
            self.logger.warning(f"Model file already exists at {self.model_paths[model_name]}")
            return

        orm_columns = self._get_orm_columns(model_name)
        orm_relationships = self._get_orm_relationships(model_name)
        if orm_relationships:
            orm_columns = orm_columns + "\n"

        content = StringUtils.replace_name_case_placeholders(MODEL_TEMPLATE, self.source_code_path, "project_name")
        content = StringUtils.replace_name_case_placeholders(content, model_name, "model_name")
        content = content.replace("{{model_columns}}", orm_columns)
        if orm_relationships:
            content = content.replace("{{model_relationships}}", orm_relationships + "\n")
        else:
            content = content.replace("{{model_relationships}}", "")
        
        self.model_content = content
    
        self.logger.info(f"Generating: {self.model_paths[model_name]}")
        self.file_service.write_file(self.model_paths[model_name], self.model_content)
        
        self.model_generated[model_name] = True
    
    def _get_migration_counter_and_path(self):
        self.logger.debug(f"_get_migration_counter_and_path()")
        versions_folder = os.path.join(self.source_code_path, "migrations", "versions")
        
        if not os.path.exists(versions_folder):
            self.logger.error(f"Error: Migrations folder not found at {versions_folder}")
            exit(1)
        
        migration_files = os.listdir(versions_folder)
        if not migration_files:
            self.migration_counter = "000001"
            self.is_default_migration_counter = True
            return self.migration_counter
        
        migration_files_counters = []
        for migration_file in migration_files:
            migration_file_counter = migration_file.split("_")[0]
            
            if not migration_file_counter.startswith("0"):
                continue
            
            migration_files_counters.append(migration_file_counter)
        
        if not migration_files_counters:
            self.migration_counter = "000001"
            self.is_default_migration_counter = True
            return self.migration_counter
        
        migration_files_counters.sort(reverse=True)
        self.migration_counter = migration_files_counters[0]
        return self.migration_counter
    
    def _increase_migration_counter(self):
        self.logger.debug(f"migration_counter: {self.migration_counter}")
        self.logger.debug(f"_increase_migration_counter()")
        new_migration_counter = str(int(self.migration_counter) + 1).zfill(self.migration_counter_digits)
        self.logger.debug(f"new_migration_counter: {new_migration_counter}")
        self.migration_path = os.path.join(self.source_code_path, "migrations", "versions", f"{new_migration_counter}_migration.py")
        self.migration_counter = new_migration_counter
        return new_migration_counter
    
    def _get_migration_columns(self, model_name):
        self.logger.debug(f"_get_migration_columns(model_name: {model_name})")
        if not self.crud_column_parsers[model_name]:
            self.logger.warning(f"No columns to migrate for {model_name}")
            return ""
        
        migration_columns = ""
        splitter = ",\n        "
        for crud_column_parser in self.crud_column_parsers[model_name]:
            migration_columns += crud_column_parser.parsed_migration_column + splitter
        return migration_columns.rstrip(splitter)
    
    def _generate_model_table_creation_statement(self, model_name):
        self.logger.debug(f"_generate_model_table_creation_statement(model_name: {model_name})")
        
        content = StringUtils.replace_name_case_placeholders(MIGRATION_MODEL_TABLE_CREATION_TEMPLATE, model_name, "model_name")
        content = content.replace("{{migration_columns}}", self._get_migration_columns(model_name))
        
        self.models_tables_creation_statements += content + "\n    "
    
    def _generate_model_table_drop_statement(self, model_name):
        self.logger.debug(f"_generate_model_table_drop_statement(model_name: {model_name})")
        
        content = StringUtils.replace_name_case_placeholders(MIGRATION_MODEL_TABLE_DROP_TEMPLATE, model_name, "model_name")
        
        self.models_tables_drop_statements += content + "\n    "
        
    def _generate_content_migration(self):
        self.logger.debug(f"_generate_content_migration()")
        
        if self.file_service.file_exists(self.migration_path):
            self.logger.warning(f"Migration file already exists at {self.migration_path}")
            return
            
        for model_name in self.model_names: 
            if not self.model_generated[model_name]:
                self.logger.warning(f"No database changes to migrate for {model_name}")
                continue    
            self._generate_model_table_creation_statement(model_name)
            self._generate_model_table_drop_statement(model_name)
            self.should_generate_migration = True
        
        if not self.should_generate_migration:
            self.logger.warning("Migration file will not be generated. No database changes to migrate")
            return
        
        self.migration_content = StringUtils.replace_name_case_placeholders(MIGRATION_TEMPLATE, self.source_code_path, "project_name")
        self.migration_content = self.migration_content.replace("{{migration_counter}}", self.migration_counter)
        self.migration_content = self.migration_content.replace("{{model_table_creation_statements}}", self.models_tables_creation_statements)
        self.migration_content = self.migration_content.replace("{{model_table_drop_statements}}", self.models_tables_drop_statements)
        
        self.logger.info(f"Generating: {self.migration_path}")
        self.file_service.write_file(self.migration_path, self.migration_content)
        self.file_service.write_json_file(self.migration_json_directory, self.schemas)
        self.file_service.write_yaml_file(self.migration_yaml_directory, self.schemas)
        
        self.migration_generated = True

    def _run_migrations(self):
        self.logger.debug(f"_run_migrations()")
        alembic_config = Config(self.alembic_config_path)
        script = ScriptDirectory.from_config(alembic_config)

        heads = script.get_heads()
        if len(heads) > 1:
            self.logger.warning(f"Multiple heads detected: {heads}")
            merge_message = "Auto-merge multiple Alembic heads"
            command.revision(alembic_config, message=merge_message, head=heads, branch_label="merge_heads")
            self.logger.info("Merged multiple heads. Now running upgrade...")
        command.upgrade(alembic_config, "head")
        
    def _generate_content_controller(self, model_name):
        self.logger.debug(f"_generate_content_controller(model_name: {model_name})")
        if self.file_service.file_exists(self.controller_paths[model_name]):
            self.logger.warning(f"Controller file already exists at {self.controller_paths[model_name]}")
            return
        
        content = StringUtils.replace_name_case_placeholders(CONTROLLER_TEMPLATE, self.source_code_path, "project_name")
        content = StringUtils.replace_name_case_placeholders(content, model_name, "model_name")
        self.controller_content = content
        
        self.logger.info(f"Generating: {self.controller_paths[model_name]}")
        self.file_service.write_file(self.controller_paths[model_name], self.controller_content)
        
        self.controller_generated[model_name] = True
    
    def _generate_content_argument_parser(self, model_name):
        self.logger.debug(f"_generate_content_argument_parser(model_name: {model_name})")
        if self.file_service.file_exists(self.argument_parser_paths[model_name]):
            self.logger.warning(f"Argument parser file already exists at {self.argument_parser_paths[model_name]}")
            return
        
        content = StringUtils.replace_name_case_placeholders(ARGUMENT_PARSER_TEMPLATE, self.source_code_path, "project_name")
        content = StringUtils.replace_name_case_placeholders(content, model_name, "model_name")
        self.argument_parser_content = content
        
        self.argument_parser_contents[model_name] = StringUtils.replace_name_case_placeholders(content, model_name, "model_name")
        self.argument_parser_contents[model_name] = StringUtils.replace_name_case_placeholders(self.argument_parser_contents[model_name], self.project_name, "project_name")
        
        self.logger.info(f"Generating: {self.argument_parser_paths[model_name]}")
        self.file_service.write_file(self.argument_parser_paths[model_name], self.argument_parser_content)
        
        self.argument_parser_generated[model_name] = True

    def _update_main_argument_parser(self, model_name):
        self.logger.debug(f"_update_main_argument_parser(model_name: {model_name})")
        
        main_argument_parser_path = os.path.join(self.source_code_path, "argument_parsers", "main_argument_parser.py")
        main_argument_parser_content = self.file_service.read_file(main_argument_parser_path)
        
        argument_parser_class_name = NamingUtils.convert_to_pascal_case(model_name) + "ArgumentParser"
        added_import_line = f"from {NamingUtils.convert_to_snake_case(self.project_name)}.argument_parsers.{self.model_names_snake_case[model_name]}_argument_parser import {argument_parser_class_name}"
        added_create_subparser_line = f"        {argument_parser_class_name}().create_subparser(subparsers)"        
        added_run_action_condition_line = f"        elif args.command == \"{model_name}\":"
        added_run_action_line = f"            {argument_parser_class_name}().run_action(args)"
        
        #### Logic starts: Go through the lines of the main_argument_parser.py file, and modify the lines based on the action
        main_argument_parser_lines = main_argument_parser_content.split("\n")
        
        added_import_index = None
        added_create_subparser_index = None
        added_run_action_index = None
        last_import_index = None
        last_create_subparser_index = None
        last_run_action_index = None
        for i, line in enumerate(main_argument_parser_lines):
            if line.startswith("from") and "import" in line:
                last_import_index = i
            if added_import_line in line:
                added_import_index = i
            if ".run_action(args)" in line:
                last_run_action_index = i
            if added_run_action_condition_line in line:
                added_run_action_index = i
            if ".create_subparser(subparsers)" in line:
                last_create_subparser_index = i
            if added_create_subparser_line in line:
                added_create_subparser_index = i
        
        if last_import_index is None and added_import_index is None:
            self.logger.error(f"Error: No import statement found in {main_argument_parser_path}")
            exit(1)
        if last_create_subparser_index is None and added_create_subparser_index is None:
            self.logger.error(f"Error: No create_subparser method found in {main_argument_parser_path}")
            exit(1)
        if last_run_action_index is None and added_run_action_index is None:
            self.logger.error(f"Error: No run_action method found in {main_argument_parser_path}")
            exit(1)

        if self.action == CrudActionsEnum.GENERATE:
            if last_import_index is not None:
                main_argument_parser_lines.insert(last_import_index + 1, added_import_line)
            if last_create_subparser_index is not None:
                main_argument_parser_lines.insert(last_create_subparser_index + 1, added_create_subparser_line)
            if last_run_action_index is not None:
                main_argument_parser_lines.insert(last_run_action_index + 1, added_run_action_condition_line)
                main_argument_parser_lines.insert(last_run_action_index + 2, added_run_action_line)
        elif self.action == CrudActionsEnum.DELETE:
            if added_import_index is not None:
                main_argument_parser_lines.remove(added_import_line)
            if added_create_subparser_index is not None:
                main_argument_parser_lines.remove(added_create_subparser_line)
            if added_run_action_index is not None:
                main_argument_parser_lines.remove(added_run_action_condition_line)
                main_argument_parser_lines.remove(added_run_action_line)
        else:
            self.logger.error(f"Unkown CrudAction: {self.action}")
            exit(1)
        
        main_argument_parser_content = "\n".join(main_argument_parser_lines)
        self.file_service.write_file(main_argument_parser_path, main_argument_parser_content)
    
    def generate(self):
        self.logger.debug(f"generate()")
        for model_name in self.model_names:
            self._generate_content_model(model_name)
            self._generate_content_controller(model_name)
            self._generate_content_argument_parser(model_name)
            
            if self.argument_parser_generated[model_name]:
                print(f"Generated: {self.argument_parser_paths[model_name]}")
            if self.controller_generated[model_name]:
                print(f"Generated: {self.controller_paths[model_name]}")
            if self.model_generated[model_name]:
                print(f"Generated: {self.model_paths[model_name]}")

            self._update_main_argument_parser(model_name)
        
        self._generate_content_migration()
        if self.migration_generated:
            print(f"Generated: {self.migration_path}")
            print(f"Generated: {self.migration_json_directory}")
            print(f"Generated: {self.migration_yaml_directory}")
        
        try:
            self._run_migrations()
            self.logger.info("Migrations ran successfully!")
        except Exception as e:
            self.logger.error(f"Error: {e}")
            self.logger.error(f"Error: {e}")
            exit(1)
         
        self.logger.info("CRUD generation completed successfully!")
        print("---")
        print("The following commands are now available:")
        for model_name in self.model_names:    
            print(f"{self.project_name} {model_name} --help | python main_code.py {model_name} --help")
            
    def delete(self):
        self.logger.debug(f"delete()")
        self.action = CrudActionsEnum.DELETE
        for model_name in self.model_names:
            if self.file_service.file_exists(self.model_paths[model_name]):
                self.file_service.delete_file(self.model_paths[model_name])
                print(f"Deleted: {self.model_paths[model_name]}")
            if self.file_service.file_exists(self.controller_paths[model_name]):
                self.file_service.delete_file(self.controller_paths[model_name])
                print(f"Deleted: {self.controller_paths[model_name]}")
            if self.file_service.file_exists(self.argument_parser_paths[model_name]):
                self.file_service.delete_file(self.argument_parser_paths[model_name])
                print(f"Deleted: {self.argument_parser_paths[model_name]}")
            self._update_main_argument_parser(model_name)
        
        print(f"All model files -except the migration file- have been deleted successfully!")
        print( "---")
        print(f"Migration file location: {os.path.dirname(self.migration_path)}")
        print(f"Make sure to modify the migration file to stop the model table creation, or delete the migration file manually if you don't want it anymore.")
        
    def delete_migration(self):
        self.logger.debug(f"delete_migration()")
        if self.file_service.file_exists(self.migration_path):
            self.file_service.delete_file(self.migration_path)
            print(f"Deleted: {self.migration_path}")





from hape.logging import Logging
from hape.hape_cli.interfaces.format_model import FormatModel

class FormatController:

    def __init__(self, format: str):
        self.logger = Logging.get_logger('hape.hape_cli.interfaces.format_controller')
        if format not in FormatModel.supported_formats:
            self.logger.error(f"Invalid format: {format}. Supported formats are: {', '.join(FormatModel.supported_formats)}")
            exit(1)
        self.format = FormatModel(format)
    
    def get(self):
        self.format.get()


import os
import json
from dotenv import load_dotenv
from hape.logging import Logging
from hape.enums.log_level_enum import LogLevelEnum
from hape.enums.environment_variables_enum import EnvironmentVariablesEnum

class Config:
    logger = Logging.get_logger('hape.config')
    _env_loaded = False
    required_env_variables = []

    _env_var_map = {
        EnvironmentVariablesEnum.HAPE_LOG_FILE: {
            "key": "HAPE_LOG_FILE",
            "value": "hape.log"
        },
        EnvironmentVariablesEnum.HAPE_LOG_LEVEL: {
            "key": "HAPE_LOG_LEVEL",
            "value": "WARNING"
        },
        EnvironmentVariablesEnum.HAPE_LOG_ROTATE_EVERY_RUN: {
            "key": "HAPE_LOG_ROTATE_EVERY_RUN",
            "value": "0"
        },
        EnvironmentVariablesEnum.HAPE_MARIADB_HOST: {
            "key": "HAPE_MARIADB_HOST",
            "value": "host.docker.internal"
        },
        EnvironmentVariablesEnum.HAPE_MARIADB_USERNAME: {
            "key": "HAPE_MARIADB_USERNAME",
            "value": "hape_user"
        },
        EnvironmentVariablesEnum.HAPE_MARIADB_PASSWORD: {
            "key": "HAPE_MARIADB_PASSWORD",
            "value": "hape_password"
        },
        EnvironmentVariablesEnum.HAPE_MARIADB_DATABASE: {
            "key": "HAPE_MARIADB_DATABASE",
            "value": "hape_db"
        },
        EnvironmentVariablesEnum.HAPE_GITLAB_DOMAIN: {
            "key": "HAPE_GITLAB_DOMAIN",
            "value": "gitlab.com"
        },
        EnvironmentVariablesEnum.HAPE_GITLAB_TOKEN: {
            "key": "HAPE_GITLAB_TOKEN",
            "value": "gitlab_token"
        }
    }

    @staticmethod
    def set_required_env_variable(required_env_variable):
        Config.logger.debug(f"set_required_env_variable({required_env_variable})")
        if required_env_variable not in Config.required_env_variables:
            Config.required_env_variables.append(required_env_variable)

    @staticmethod
    def set_env_var_key(hape_key, new_key):
        for key, _ in Config._env_var_map.items():
            if key == hape_key:
                Config._env_var_map[hape_key]["key"] = new_key

    @staticmethod
    def check_variables():
        Config.logger.debug(f"check_variables({Config.required_env_variables})")
        for variable in Config.required_env_variables:
            Config._get_env_value(variable)

    @staticmethod
    def _load_environment():
        if not Config._env_loaded:
            if os.path.exists(".env"):
                load_dotenv()
            Config._env_loaded = True

    @staticmethod
    def _get_env_value(hape_env_key):
        Config._load_environment()
        env_key = Config._env_var_map[hape_env_key]["key"]
        env_value = os.getenv(env_key)
        env_default_value = Config._env_var_map[hape_env_key]["value"]
        
        if env_value:
            Config._env_var_map[hape_env_key]["value"] = env_value
            return env_value
        elif env_default_value:
            return env_default_value
        elif env_key in Config.required_env_variables:
            Config.logger.error(f"""Environment variable {env_key} is missing.

To set the value of the environment variable run:
$ export {env_key}="value"

The following environment variables are required:
{json.dumps(Config.required_env_variables, indent=4)}
""")
            exit(1)    
        return None
    
    @staticmethod
    def get_log_level():
        Config.logger.debug(f"get_log_level()")
        log_level = Config._get_env_value(EnvironmentVariablesEnum.HAPE_LOG_LEVEL)
        return log_level if log_level and LogLevelEnum(log_level) else LogLevelEnum.WARNING

    @staticmethod
    def get_log_file():
        Config.logger.debug(f"get_log_file()")
        log_file = Config._get_env_value(EnvironmentVariablesEnum.HAPE_LOG_FILE)
        return log_file if log_file else "hape.log"
    
    @staticmethod
    def get_log_rotate_every_run():
        Config.logger.debug(f"get_log_rotate_every_run()")
        log_rotate_every_run = Config._get_env_value(EnvironmentVariablesEnum.HAPE_LOG_ROTATE_EVERY_RUN)
        return log_rotate_every_run if log_rotate_every_run else "0"



from hape.logging import Logging
from hape.base.model_argument_parser import ModelArgumentParser
from hape.models.k8s_deployment_cost_model import K8SDeploymentCost
from hape.controllers.k8s_deployment_cost_controller import K8SDeploymentCostController

class K8SDeploymentCostArgumentParser(ModelArgumentParser):
    def __init__(self):
        super().__init__(K8SDeploymentCost, K8SDeploymentCostController)
        self.logger = Logging.get_logger('hape.argument_parsers.k8s_deployment_cost_argument_parser')

    def extend_subparser(self):
        pass
    
    def extend_actions(self):
        pass


from setuptools import setup, find_packages

setup(
    name="hape",
    version="0.2.113",
    packages=["hape"],
    include_package_data=True,
    install_requires=[
        "alembic==1.14.1",
        "build==1.2.2.post1",
        "cachetools==5.5.1",
        "certifi==2024.12.14",
        "cffi==1.17.1",
        "charset-normalizer==3.4.1",
        "cryptography==44.0.0",
        "docutils==0.21.2",
        "durationpy==0.9",
        "google-auth==2.38.0",
        "greenlet==3.1.1",
        "idna==3.10",
        "iniconfig==2.0.0",
        "jaraco.classes==3.4.0",
        "jaraco.context==6.0.1",
        "jaraco.functools==4.1.0",
        "keyring==25.6.0",
        "kubernetes==31.0.0",
        "Mako==1.3.8",
        "markdown-it-py==3.0.0",
        "MarkupSafe==3.0.2",
        "mdurl==0.1.2",
        "more-itertools==10.6.0",
        "mysql==0.0.3",
        "mysql-connector-python==9.2.0",
        "mysqlclient==2.2.7",
        "nh3==0.2.20",
        "oauthlib==3.2.2",
        "packaging==24.2",
        "pkginfo==1.12.0",
        "pluggy==1.5.0",
        "pyasn1==0.6.1",
        "pyasn1_modules==0.4.1",
        "pycparser==2.22",
        "Pygments==2.19.1",
        "PyMySQL==1.1.1",
        "pyproject_hooks==1.2.0",
        "pytest==8.3.4",
        "python-dateutil==2.9.0.post0",
        "python-dotenv==1.0.1",
        "python-gitlab==5.6.0",
        "python-json-logger==3.2.1",
        "PyYAML==6.0.2",
        "readme_renderer==44.0",
        "requests==2.32.3",
        "requests-oauthlib==2.0.0",
        "requests-toolbelt==1.0.0",
        "rfc3986==2.0.0",
        "rich==13.9.4",
        "rsa==4.9",
        "ruamel.yaml==0.18.10",
        "ruamel.yaml.clib==0.2.12",
        "setuptools==75.8.0",
        "six==1.17.0",
        "SQLAlchemy==2.0.37",
        "twine==6.0.1",
        "typing_extensions==4.12.2",
        "urllib3==2.3.0",
        "websocket-client==1.8.0",
        "wheel==0.45.1",
    ],
    entry_points={
        "console_scripts": [
            "hape=hape.hape_cli.cli:main",
        ],
    },
    author="Hazem Ataya",
    author_email="hazem.ataya94@gmail.com",
    description="HAPE Framework: Build an Automation Tool With Ease",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/hazemataya94/hape-framework",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



from hape.logging import Logging
from hape.hape_cli.interfaces.format_argument_parser import FormatArgumentParser

class YamlArgumentParser(FormatArgumentParser):
    def __init__(self):
        super().__init__("yaml")
        self.logger = Logging.get_logger('hape.hape_cli.argument_parsers.yaml_argument_parser')



from hape.logging import Logging
from hape.base.model_controller import ModelController
from hape.models.k8s_deployment_model import K8SDeployment

class K8SDeploymentController(ModelController):
    
    def __init__(self):
        super().__init__(K8SDeployment)
        self.logger = Logging.get_logger('hape.controllers.k8s_deployment_controller')


from hape.hape_cli.init_templates.source_code_templates import BOOTSTRAP_PY, MAIN_PY, PLAYGROUND_PY, CLI_PY, MAIN_ARGUMENT_PARSER, PLAYGROUND_ARGUMENT_PARSER, MIGRATION_ENV_PY
from hape.hape_cli.init_templates.dockerfiles_templates import DOCKERFILE_DEV, DOCKERFILE_PROD, DOCKER_COMPOSE
from hape.hape_cli.init_templates.hidden_files_templates import DOCKER_IGNORE, ENV_EXAMPLE, GIT_IGNORE
from hape.hape_cli.init_templates.requirements_templates import REQUIREMENTS_DEV, REQUIREMENTS_CLI
from hape.hape_cli.init_templates.makefile_template import MAKEFILE
from hape.hape_cli.init_templates.alembic_template import ALEMBIC_INI
from hape.hape_cli.init_templates.setup_templates import SETUP_PY, MANIFEST_IN

INIT_PROJECT_TEMPLATE = {
    ".dockerignore": DOCKER_IGNORE,
    ".env.example": ENV_EXAMPLE,
    ".gitignore": GIT_IGNORE,
    "alembic.ini": ALEMBIC_INI,
    "main.py": MAIN_PY,
    "Makefile": MAKEFILE,
    "MANIFEST.in": MANIFEST_IN,
    "README.md": "# {{project_name_title_case}}",
    "requirements-dev.txt": REQUIREMENTS_DEV,
    "requirements-cli.txt": REQUIREMENTS_CLI,
    "setup.py": SETUP_PY,
    "dockerfiles": {
        "Dockerfile.dev": DOCKERFILE_DEV,
        "Dockerfile.prod": DOCKERFILE_PROD,
        "docker-compose.yml": DOCKER_COMPOSE,
    },
    "{{project_name_snake_case}}": {
        "__init__.py": None,
        "cli.py": CLI_PY,
        "bootstrap.py": BOOTSTRAP_PY,
        "playground.py": PLAYGROUND_PY,
        "argument_parsers": {
            "__init__.py": None,
            "playground_argument_parser.py": PLAYGROUND_ARGUMENT_PARSER,
            "main_argument_parser.py": MAIN_ARGUMENT_PARSER
        },
        "controllers": {
            "__init__.py": None   
        },
        "enums": {
            "__init__.py": None
        },
        "migrations": {
            "README": None,
            "env.py": MIGRATION_ENV_PY,
            "script.py.mako": None,
            "versions": {
                ".gitkeep": None
            },
            "json": {
                ".gitkeep": None
            },
            "yaml": {
                ".gitkeep": None
            }
        },
        "models": {
            "__init__.py" : None  
        },
        "services": {
            "__init__.py": None
        },
    }
}


from typing import List
from hape.utils.naming_utils import NamingUtils
from hape.logging import Logging
from hape.hape_cli.models.crud_column import CrudColumn
from hape.hape_cli.enums.crud_column_valid_properties import CrudColumnValidPropertiesEnum
from hape.hape_cli.enums.crud_column_valid_types import CrudColumnValidTypesEnum
from hape.hape_cli.enums.crud_column_fk_on_delete import CrudColumnFkOnDeleteEnum


class CrudColumnParser:
    
    def __init__(self, crud_column: CrudColumn):
        self.logger = Logging.get_logger('hape.hape_cli.models.crud_column_parser')
        self.crud_column = crud_column
        
        self.orm_column_name = NamingUtils.convert_to_snake_case(self.crud_column.name)
        self.orm_column_type = self._parse_orm_column_type()
        self.orm_relationships = "" # set in _parse_orm_column_properties()
        self.orm_column_properties = self._parse_orm_column_properties()
        
        self.parsed_orm_column_template = "{{model_column_name_snake_case}} = Column({{orm_column_type}}, {{orm_column_properties}})"
        self.parsed_orm_column = self.parsed_orm_column_template.replace("{{model_column_name_snake_case}}", self.orm_column_name)
        self.parsed_orm_column = self.parsed_orm_column.replace("{{orm_column_type}}", self.orm_column_type)
        self.parsed_orm_column = self.parsed_orm_column.replace("{{orm_column_properties}}", self.orm_column_properties)
        self.parsed_orm_column = self.parsed_orm_column.replace(", )", ")")

        self.parsed_migration_column_template = "sa.Column('{{model_column_name_snake_case}}', sa.{{orm_column_type_snake_case}}, {{orm_column_properties}})"
        self.parsed_migration_column = self.parsed_migration_column_template.replace("{{model_column_name_snake_case}}", self.orm_column_name)
        self.parsed_migration_column = self.parsed_migration_column.replace("{{orm_column_type_snake_case}}", self.orm_column_type)
        self.parsed_migration_column = self.parsed_migration_column.replace("{{orm_column_properties}}", self.orm_column_properties)
        self.parsed_migration_column = self.parsed_migration_column.replace(", )", ")")
        self.parsed_migration_column = self.parsed_migration_column.replace("ForeignKey", "sa.ForeignKey")
        
    def _parse_orm_column_type(self):
        self.logger.debug(f"_parse_orm_column_type()")
        
        orm_column_type = ""
        
        if self.crud_column.type == CrudColumnValidTypesEnum.INT:
            orm_column_type = "Integer"
        elif self.crud_column.type == CrudColumnValidTypesEnum.FLOAT:
            orm_column_type = "Float"
        elif self.crud_column.type == CrudColumnValidTypesEnum.BOOL:
            orm_column_type = "Boolean"
        elif self.crud_column.type == CrudColumnValidTypesEnum.DATE:
            orm_column_type = "Date"
        elif self.crud_column.type == CrudColumnValidTypesEnum.DATETIME:
            orm_column_type = "DateTime"
        elif self.crud_column.type == CrudColumnValidTypesEnum.TIMESTAMP:
            orm_column_type = "TIMESTAMP"
        elif self.crud_column.type == CrudColumnValidTypesEnum.TEXT:
            orm_column_type = "Text"
        elif self.crud_column.type == CrudColumnValidTypesEnum.STRING:
            orm_column_type = "String(255)"
        else:
            self.logger.error(f"Invalid column type: {self.crud_column.type}")
            exit(1)
                
        return orm_column_type
                
    def _parse_orm_column_properties(self):
        self.logger.debug(f"_parse_orm_column_properties()")
        orm_column_properties = ""
        orm_column_relationships = ""
        for property in self.crud_column.crud_column_properties:
            if property.property == CrudColumnValidPropertiesEnum.PRIMARY:
                orm_column_properties += "primary_key=True, "
            elif property.property == CrudColumnValidPropertiesEnum.AUTOINCREMENT:
                orm_column_properties += "autoincrement=True, "
            elif property.property == CrudColumnValidPropertiesEnum.UNIQUE:
                orm_column_properties += "unique=True, "
            elif property.property == CrudColumnValidPropertiesEnum.INDEX:
                orm_column_properties += "index=True, "
            elif property.property == CrudColumnValidPropertiesEnum.FOREIGN_KEY:
                orm_column_properties += f"ForeignKey('{property.foreign_key.foreign_key_table}.{property.foreign_key.foreign_key_column}', ondelete='{property.foreign_key.foreign_key_on_delete}'), "
                
                foreign_key_model_pascal_case = NamingUtils.convert_to_pascal_case(property.foreign_key.foreign_key_table)
                
                orm_column_relationships += f"relationship('{foreign_key_model_pascal_case}', back_populates='{self.orm_column_name.rstrip('_id').rstrip('s')}s')\n    "
                
        self.orm_relationships = orm_column_relationships.strip()
        
        for property in self.crud_column.crud_column_properties:
            
            if property.property == CrudColumnValidPropertiesEnum.NULLABLE:
                orm_column_properties += "nullable=True"
            elif property.property == CrudColumnValidPropertiesEnum.REQUIRED:
                orm_column_properties += "nullable=False"
                
        return orm_column_properties
    


import enum

class CrudColumnValidTypesEnum(enum.Enum):
    STRING = "string"
    TEXT = "text"
    INT = "int"
    BOOL = "bool"
    FLOAT = "float"
    DATE = "date"
    DATETIME = "datetime"
    TIMESTAMP = "timestamp"



import os
import sys
import shutil
import csv
import json
from ruamel.yaml import YAML
from hape.logging import Logging

class FileService:

    def __init__(self):
        self.yaml = YAML()
        self.yaml.width = sys.maxsize
        self.yaml.preserve_quotes = True
        self.yaml.indent(mapping=2, sequence=4, offset=2)
        self.logger = Logging.get_logger('hape.services.file_service')
        
    def create_directory(self, path: str):
        self.logger.debug(f"create_directory(path: {path})")
        os.makedirs(path, exist_ok=True)
    
    def create_directory(self, path: str):
        self.logger.debug(f"create_directory(path: {path})")
        os.makedirs(path, exist_ok=True)

    def delete_file(self, file_path: str):
        self.logger.debug(f"delete_file(file_path: {file_path})")
        if os.path.exists(file_path):
            os.remove(file_path)
        self.logger.info(f"File '{file_path}' has been removed.")  
        
    def delete_folder(self, folder_path: str):
        self.logger.debug(f"delete_folder(folder_path: {folder_path})")
        self.delete_directory(folder_path)

    def delete_directory(self, directory_path: str):
        self.logger.debug(f"delete_directory(directory_path: {directory_path})")
        if os.path.exists(directory_path) and os.path.isdir(directory_path):
            try:
                shutil.rmtree(directory_path)
                self.logger.info(f"Directory '{directory_path}' has been removed.")
            except Exception as e:
                self.logger.error(f"Unable to remove directory '{directory_path}'. {e}")
        else:
            self.logger.warning(f"Directory '{directory_path}' does not exist or is not a directory.")

    def copy_file(self, source: str, destination: str, overwrite: bool):
        self.logger.debug(f"copy_file(source: {source}, destination: {destination}, overwrite: {overwrite})")
        if os.path.exists(source) or overwrite:
            shutil.copy2(source, destination)

    def copy_directory(self, source: str, destination: str):
        self.logger.debug(f"copy_directory(source: {source}, destination: {destination})")
        if os.path.exists(source):
            self.logger.info(f"copying {source} to {destination}")
            shutil.copytree(source, destination, dirs_exist_ok=True)
    
    def file_exists(self, path):
        self.logger.debug(f"path_exists(path: {path})")
        return os.path.exists(path)
    
    def folder_exists(self, path):
        self.logger.debug(f"path_exists(path: {path})")
        return os.path.exists(path)

    def directory_exists(self, path):
        self.logger.debug(f"path_exists(path: {path})")
        return os.path.exists(path)
        
    def path_exists(self, path):
        self.logger.debug(f"path_exists(path: {path})")
        return os.path.exists(path)

    def replace_text_in_file(self, source: str, destination: str, old_text: str, new_text: str):
        self.logger.debug(f"replace_text_in_file(source {source}, destination {destination}, old_text {old_text}, new_text {new_text})")
        if os.path.exists(source):
            with open(source, "r", encoding="utf-8") as src, open(destination, "w", encoding="utf-8") as dest:
                content = src.read().replace(old_text, new_text)
                dest.write(content)

    def write_file(self, file_path, content):
        self.logger.debug(f"write_file(file_path: {file_path}, content)")
        with open(file_path, 'w', encoding='utf-8') as file:
            file.write(content)

    def read_file(self, file_path):
        self.logger.debug(f"read_file(file_path: {file_path})")
        if not os.path.exists(file_path):
            self.logger.error(f"Error: file {file_path} does not exist")
            exit(1)
        with open(file_path, 'r', encoding='utf-8') as source_file:
            content = source_file.read()
        return content

    def read_yaml_file(self, yaml_path):
        self.logger.debug(f"read_yaml_file(yaml_path: {yaml_path})")
        if not os.path.exists(yaml_path):
            self.logger.error(f"Error: file {yaml_path} does not exist")
            exit(1)
        with open(yaml_path, 'r', encoding='utf-8') as file:
            data = self.yaml.load(file)
        return data
    
    def read_json_file(self, json_path):
        self.logger.debug(f"read_json_file(json_path: {json_path})")
        if not os.path.exists(json_path):
            self.logger.error(f"Error: file {json_path} does not exist")
            exit(1)
        with open(json_path, 'r', encoding='utf-8') as file:
            try:
                data = json.load(file)
            except json.JSONDecodeError as e:
                self.logger.error(f"Error: file {json_path} is not a valid JSON file. {e}")
                exit(1)
        return data

    def write_yaml_file(self, yaml_path, data):
        self.logger.debug(f"write_yaml_file(yaml_path: {yaml_path}, data)")
        with open(yaml_path, 'w') as file:
            self.yaml.dump(data, file)

    def write_json_file(self, json_path, data):
        self.logger.debug(f"write_json_file(json_path: {json_path}, data)")
        with open(json_path, 'w') as file:
            data = json.dumps(data, indent=4)
            file.write(data)

    def append_to_file(self, file_path, content):
        self.logger.debug(f"append_to_file(file_path: {file_path}")
        if not os.path.exists(file_path):
            self.logger.error(f"Error: file {file_path} does not exist")
            exit(1)
        with open(file_path, 'a', encoding='utf-8') as destination_file:
            destination_file.write(content)

    def prepend_to_file(self, file_path, content):
        self.logger.debug(f"prepend_to_file(file_path: {file_path}")
        if not os.path.exists(file_path):
            self.logger.error(f"Error: file {file_path} does not exist")
            exit(1)
        with open(file_path, 'r', encoding='utf-8') as file:
            file_content = file.read()
        with open(file_path, 'w', encoding='utf-8') as file:
            file.write(content + '\n' + file_content)

    def add_string_after_keyword(self, file_path, keyword, string_to_add='\n'):
        self.logger.debug(f"add_string_after_keyword(file_path: {file_path}, keyword: {keyword}, string_to_add: {string_to_add})")
        if not os.path.exists(file_path):
            self.logger.error(f"Error: file {file_path} does not exist")
            exit(1)
        with open(file_path, 'r+', encoding='utf-8') as file:
            content = file.readlines()
            for i, line in enumerate(content):
                if keyword in line:
                    content.insert(i + 1, string_to_add + '\n')
                    break
            file.seek(0)
            file.writelines(content)
            file.truncate()

    def read_csv_file(self, csv_path):
        self.logger.debug(f"read_csv_file(csv_path: {csv_path})")
        if not os.path.exists(csv_path):
            self.logger.error(f"Error: file {csv_path} does not exist")
            exit(1)
        with open(csv_path, 'r', encoding='utf-8') as csvfile:
            reader = csv.DictReader(csvfile)
            clean_data = []
            for row in reader:
                clean_row = {key.strip(): value.strip() for key, value in row.items()}
                clean_data.append(clean_row)
            return clean_data

    def write_csv_file(self, filename, data):
        self.logger.debug(f"write_csv_file(filename: {filename}, data)")
        if not data:
            self.logger.info("No data provided to write CSV file.")
            return
        fieldnames = list(data[0].keys())
        with open(filename, mode="w", newline="", encoding="utf-8") as file:
            writer = csv.DictWriter(file, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerows(data)

    def find_files_with_keyword(self, keyword, directory, return_parent_directory=False):
        self.logger.debug(f"find_files_with_keyword(keyword: {keyword}, directory: {directory}, return_parent_directory: {return_parent_directory})")
        matching_files = []
        for root, dirs, files in os.walk(directory):
            for file in files:
                if keyword in file:
                    if return_parent_directory:
                        matching_files.append(os.path.dirname(os.path.join(root, file)))
                    else:
                        matching_files.append(os.path.join(root, file))
        return matching_files

    def get_sorted_subdirectories(self, dir_path, prefix):
        self.logger.debug(f"get_sorted_subdirectories(dir_path: {dir_path}, prefix: {prefix})")
        subdirectories = sorted(os.listdir(dir_path))
        return [os.path.join(dir_path, subdir) for subdir in subdirectories
                if subdir.startswith(prefix) and os.path.isdir(os.path.join(dir_path, subdir))]



import json
import logging.config
from hape.logging import Logging, CustomJsonFormatter, LOGGING_CONFIG
from hape.config import Config

def bootstrap():
    log_level = Config.get_log_level()
    log_file = Config.get_log_file()
    
    if Config.get_log_rotate_every_run() == "1":
        Logging.rotate_log_file(log_file)
    
    logging_config_json = LOGGING_CONFIG.replace("{{log_level}}", log_level).replace("{{log_file}}", log_file)
    logging_config = json.loads(logging_config_json)
    logging_config["formatters"]["json"] = {
        "()": CustomJsonFormatter,
        "fmt": "%(timestamp)s %(level)s %(name)s %(message)s %(module)s %(funcName)s %(lineno)d"
    }
    logging.config.dictConfig(logging_config)
    logger = Logging.get_logger()
    logger.info("Application started!")



DOCKER_IGNORE = """
.venv
dockerfiles
**docker-compose.yml**
*.md
""".strip()

ENV_EXAMPLE = """
{{project_name_upper_snake_case}}_MARIADB_HOST="host.docker.internal"
{{project_name_upper_snake_case}}_MARIADB_USERNAME="{{project_name_snake_case}}_user"
{{project_name_upper_snake_case}}_MARIADB_PASSWORD="{{project_name_snake_case}}_password"
{{project_name_upper_snake_case}}_MARIADB_DATABASE="{{project_name_snake_case}}_db"
""".strip()

GIT_IGNORE = """
dockerfiles/*-init
playground/*
!playground/.gitkeep
!{{project_name_snake_case}}/migrations/.gitkeep
*.zip
*.venv
*.vscode

.DS_Store
*.DS_Store
**.DS_Store

# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# C extensions
*.so

# Distribution / packaging
.Python
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
wheels/
pip-wheel-metadata/
share/python-wheels/
*.egg-info/
.installed.cfg
*.egg

# PyInstaller
# Usually these files are written by a python script from a template
# before PyInstaller builds the executable, when PyInstaller is instructed to
# follow the import statements of scripts it is given, and so they must be
# included in source control in order to be able to recreate the executable
# later from this source control.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit test / coverage reports
htmlcov/
.tox/
.nox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*.cover
*.py,cover
.hypothesis/
.pytest_cache/

# Translations
*.mo
*.pot

# Django stuff:
*.log
local_settings.py
db.sqlite3
db.sqlite3-journal

# PyBuilder
target/

# Jupyter Notebook
.ipynb_checkpoints

# IPython
profile_default/
ipython_config.py

# pyenv
.python-version

# Environments
.env
.venv
venv/
""".strip()



CONTROLLER_TEMPLATE = """
from hape.logging import Logging
from hape.base.model_controller import ModelController
from {{project_name_snake_case}}.models.{{model_name_snake_case}}_model import {{model_name_pascal_case}}

class {{model_name_pascal_case}}Controller(ModelController):
    
    def __init__(self):
        super().__init__({{model_name_pascal_case}})
        self.logger = Logging.get_logger('{{project_name_snake_case}}.controllers.{{model_name_snake_case}}_controller')
""".strip()


from hape.logging import Logging

class ModelController():

    def __init__(self, base_model_class):
        self.logger = Logging.get_logger('hape.base.model_controller')
        self._base_model_class = base_model_class
    
    def save(self, base_model):
        self.logger.debug(f"save()")
        if base_model.validate():
            try:
                if base_model.save():
                    return True
                raise RuntimeError(f"Couldn't save {base_model.json()}")
            except Exception as e:
                raise RuntimeError(f"An error occurred while saving: {e}")
        
        raise ValueError("Model validation failed.")

    def get(self, **filters):
        self.logger.debug(f"get({filters})")
        if not filters:
            raise ValueError("Filters cannot be empty for get operation.")
        return self._base_model_class.get(**filters)
    
    def get_all(self, **filters):
        self.logger.debug(f"get_all({filters})")
        return self._base_model_class.get_all(**filters)
    
    def delete(self, base_model):
        self.logger.debug(f"delete()")
        if not isinstance(base_model, self._base_model_class):
            raise TypeError("Invalid base_model instance for deletion.")
        return base_model.delete()
    
    def delete_all(self, **filters):
        self.logger.debug(f"delete_all({filters})")
        if not filters:
            raise ValueError("Filters cannot be empty for delete_all operation.")
        return self._base_model_class.delete_all(**filters)
